print("import: 'transformers'")
import transformers

print("import: 'transformers.benchmark'")
import transformers.benchmark

print("import: 'transformers.commands'")
import transformers.commands

print("import: 'transformers.data'")
import transformers.data

print("import: 'transformers.kernels'")
import transformers.kernels

print("import: 'transformers.models'")
import transformers.models

print("import: 'transformers.pipelines'")
import transformers.pipelines

print("import: 'transformers.tools'")
import transformers.tools

print("import: 'transformers.utils'")
import transformers.utils

